/* algnum_mat.gp */
        
/* Matrices of algebraic numbers are represented as a primitive element of
   a field containing all the entries and a matrix of polymods; this allows
   simple lifting of many operations. */
a.prim = a[1]
a.data = a[2]

anmatinit(a) =
{
  local(m, n, v);
  
  if(type(a)=="t_VEC", return(anvecinit(a)));
  
  m = matsize(a)[1]; n = matsize(a)[2]; 
  v = anvecinit(vector(m*n, i, a[(i-1)\n+1, (i-1)%n+1]));
  [v.prim, matrix(m,n, i,j, v.data[(i-1)*n+j])]  
}

anmatapprox(a) =
{ 
  if(type(a.data) == "t_MAT",
    matrix(matsize(a.data)[1], matsize(a.data)[2], i, j,
           subst(lift(a.data[i,j]), x, a.prim.approx)),
    vector(#a.data, i, subst(lift(a.data[i]), x, a.prim.approx)))
}

anvecinit(a) =
{
  local(s = length(a), u = anrat(1), m = vector(s), c);
  
  for(i=1, s,
    c = ancompositum(u, a[i],1);
    u = c[1]; m[i] = lift(c[3]);
    for(p=1, i-1,
      m[p] = lift(subst(m[p], x, c[2]))));
  for(i=1, s,
    m[i] = Mod(m[i], u.minpol));
  [u,m]
}

anmatdet(m) = ansubst(matdet(m[2]), x, m[1])

embed(a, p) =
{
  local(v);

  if(type(a) == "t_MAT",
    matrix(matsize(a)[1], matsize(a)[2], i, j,
           subst(lift(a[i,j]), x, p)),
    v = vector(#a, i, subst(lift(a[i]), x, p));
    if(type(a) == "t_VEC", v, Col(v)))
}

anmatmul(a,b) =
{
  local(c = ancompositum(a.prim, b.prim,1));
  [c[1], embed(a.data, c[2]) * embed(b.data, c[3])]
}

anmatadd(a,b) =
{
  local(c = ancompositum(a[1], b[1],1));
  [c[1], embed(a.data, c[2]) + embed(b.data, c[3])]
}

anmattranspose(a) = [a.prim, a.data~]

