/* arithmeticity.gp */


arithmetic2(G, itf=1, qa=1) =
{
  local(t);
  
  if(itf==1, itf = invariant_trace_field(G));
  if(qa==1, qa = invariant_quaternion_algebra(G));
  
  t = nfinit(itf.minpol).r2; if(t!=1, return(Str("Invariant trace field has ",t, " complex places")));
  t = allint(G); if(t!=1, return(t));
  t = real_ramification(qa);
  for(i=1, #t, if(t[i]==0, return(Str(i, "th root unramified"))));
  return(1);
}
  
arithmetic1(G,k=1,K=1) =
{
  local(n = matsize(G)[1], t);
  
  if(k==1, k = littlek(G));
  if(K==1, K = extend_field(anrat(0), G));
  if(debug,print(k));
  t = nfinit(K.minpol).r2; if(t>0, return(Str("K(P) field has ",t, " complex places")));
  t = allint(G); if(t!=1, return(t));
  if(debug,print("checking positive definite embeddings."));
  t = positive_semidefinite_embeddings(mat_in_terms_of(G,K),k,K); return(t)
}

allint(m) =
{
  local(s = matsize(m));
  for(i=1,s[1],
    for(j=1,s[2],
      if(abs(polcoeff(m[i,j].minpol, andegree(m[i,j]))) != abs(content(m[i,j].minpol)),
        return(Str("G[", i, ",", j, "] is not an algebraic integer")))));
  1
}

in_terms_of(u, v) =
{
  local(c = ancompositum(u, v, 1));
  return(subst(lift(c[2]), x, modreverse(c[3])))
}

mat_in_terms_of(M, v) =
{
  local(s = matsize(M));
  return(matrix(s[1], s[2], i, j, in_terms_of(M[i,j], v)))
}

positive_semidefinite_embeddings(G_K,k,K) =
{
  local(n = matsize(G)[1],
        K_conjugates = polroots(K.minpol),
        k_K = in_terms_of(k, K),
        K_image,k_image, G_image);
  
  for(m=1, andegree(K),
    K_image = aninit(K.minpol,K_conjugates[m],0);
    k_image = ansubst(k_K, x,K_image);
    if(debug,print("k_image: ",k_image));
    if(!aneq(k, k_image),
      if(debug,print("Moving littlek."));
      G_image = [K_image,G_K];
      if(!positive_semidefinite(G_image),
        return(Str("embedding #", m, " does not give a ",
                   "positive semi-definite image for G")))));
  return(1) 
}

positive_semidefinite(M) =
{
  local(M_minor,M_minor_det,zero =anrat(0));
  for(i=1, matsize(M.data)[1],
   M_minor = [M.prim,matrix(i,i,j,k,M.data[j,k])];
   M_minor_det = anmatdet(M_minor);
   if(debug,print(M_minor_det));
   if(!aneq(M_minor_det,zero) && real(M_minor_det.approx) < 0,
     return(0)));
  return(1)
}
