/* field_prim */

generate_same_field(a,b) = 
{
  local(c = ancompositum(a,b));
  andegree(c) == andegree(a) && andegree(c) == andegree(b)
}

/*takes the field generated by algebraic number a and provides a reduced
 * representation via a vector containing [b,c], with
an algebraic number b having minimal polynomial that is simplified by polredabs
and c expressing the root of a.minpol as a polymod modulo b.minpol. */
reducefield(a,canonical = 0,old_root_in_new = 0) =
{
  local(p = a[1],oldroot = a[2],
        reduced = if(canonical, canonical_minpoly(p),polredabs(p,1)),
        q = reduced[1],
        polsubst = lift(modreverse(reduced[2])),
        newroot = subst(polsubst,x,oldroot),
        newprim = aninit(q, newroot));
        
  if(old_root_in_new, [newprim,reduced[2]], newprim);
}
addhelp(reducefield, "reducefield(a): provides an algebraic number b so "\
        "that Q(b) = Q(a), but so that b.minpoly = polredabs(a.minpoly)")

/*  A good polynomial to check this on: x^12-4*x^9+22*x^6-12*x^3+2 */
/* or, if you are very ambitious: x^16-4*x^15+8*x^14-8*x^13+6*x^12-16*x^11+40*x^10-32*x^9-50*x^8+160*x^7-176*x^6+40*x^5+140*x^4-192*x^3+112*x^2-32*x+4  which has very many returned polynomials from polredabs. */

/* Used in canonical_minpoly */
pol_denominator(p) = 
{
  local(deg, c, my_primes, number_of_primes, coef_factor, required_power);
  
  deg = poldegree(p);
  c = content(p); /*This is the GCD of the coefficients of p(x)...
                          in particular it provides the LCM of the denominators */
  if(c == 1, return(c));
  my_primes = factor(c);
  number_of_primes = matsize(my_primes)[1];

  /*clear the multiplicities of the prime factors, we will use this 
    for a different meaning */
  for(i=1, number_of_primes, my_primes[i,2] = 0);

  for(coefficient=0, deg-1,
    coeff_factor = factor(polcoeff(p,coefficient));
    n = matsize(coeff_factor)[1];

    i = 1;  /* index of the prime within "my_primes" */
    for (j=1,n, /*index of the prime within the factorization of
                  the current coefficient*/
      while(i <= number_of_primes && my_primes[i,1] < coeff_factor[j,1], i++);
      if(i > number_of_primes, break());
      required_power = ceil(-coeff_factor[j,2]/(deg-coefficient));
      if (my_primes[i,2] < required_power, my_primes[i,2] = required_power)));

  my_denominator = 1;
  for (i=1, number_of_primes,
    my_denominator *= my_primes[i,1]^my_primes[i,2]);
  return(my_denominator);
}
          

canonical_minpoly(p) = 
{
  local(deg, denom, candidates, num, fix_sign, which, matrix_for_sorting,
        lex_vector, newpoly, old_root_in_terms_of_new);
  
  deg = poldegree(p);

  /* if p has any rational coefficients, we need to remove them by the following
  change of variables. The roots of the new polynomial are the roots of the
  previous one multiplied by denom. */

  denom = pol_denominator(p);
  p = denom^deg * subst(p,x,x/denom);

  candidates = polredabs(p,5);
  num = length(candidates);
  fix_sign = vector(num);  /*Keeps track of which polynomials from this
                             polredabs we have modified so that they
		             have minus the specified root */
  which = 0;

  if(verbose, print(candidates, " "));

  /* The following makes a chance of the signs of coefficients in order to
  enforce the condition that the first non-zero coefficient congruent with deg-1
  mod(2) when decreasing from deg-1 is negative. Remaining coefficients of the
  same parity are also modified so that the field generated by the polynomial
  remains unchanged. IT MAY BE THE CASE THAT THIS IS ALREADY DONE IN POLREDABS
  This appears in the Snap code, so we repeat it.*/

  /* FOR LATER---USE subst with -x */
  for (j=1,num,
    pol = Vec(candidates[j][1]);
    fix_sign[j] = 1;
    i = 2;  /* start with the coefficient of degree deg -1  */
    while (i <= deg+1 && pol[i] == 0, i = i + 2); 
    if (i <= deg+1 && pol[i] > 0,
      while (i < deg,pol[i] = -pol[i]; i = i + 2);
      fix_sign[j] = -1;
      candidates[j][1] = Pol(pol)));
     
  if(verbose, print(candidates, " "));
  
  matrix_for_sorting = matrix(num,deg+4);

  for(i=1,num,
   matrix_for_sorting[i,1] = abs(poldisc(candidates[i][1]));
   matrix_for_sorting[i,deg+4] = i;
   power_of_two = 1;
   for (j=0,deg,
     matrix_for_sorting[i,j+2] = abs(polcoeff(candidates[i][1],deg-j));
     matrix_for_sorting[i,deg+3] =  matrix_for_sorting[i,deg+3] + sign(polcoeff(candidates[i][1],deg-j))*power_of_two;
     power_of_two = power_of_two*2));

  if(verbose, print(" ", matrix_for_sorting));

  lex_vector = vector(deg+3,i,i);
  
  sorted_matrix = vecsort(matrix_for_sorting~,lex_vector,2)~;
  if(verbose, print("  ",sorted_matrix));
  which = sorted_matrix[1,deg+4];

  newpoly = candidates[which];

  if(verbose, print(newpoly));
  old_root_in_terms_of_new = (candidates[which][2]/denom)*fix_sign[which];
  newpoly[2] = old_root_in_terms_of_new; /*Expressing the roots of the reduced polynomial in terms of the initial one.*/

  newpoly
}
      


