/* quaterion_algebra.gp */

/*Hilbert symbols have 3 pieces of member data:*/
x.prim = x[1];   /*A primitive element for the invariant trace field */
x.first = x[2];   /*first and second are polymods with modulus x.field.minpol */
x.second = x[3];

hilbert_symbol_init(field,first,second,canonical = 1) = 
{
  local(data, reduced_prim, new_root_on_old, reduced_first, reduced_second);
  data = reducefield(field,canonical,1);
  if(verbose, print(data));
  reduced_prim = data[1];
  new_root_on_old = data[2];
  reduced_first = Mod(subst(lift(first),x,lift(new_root_on_old)),reduced_prim.minpol);
  reduced_second = Mod(subst(lift(second),x,lift(new_root_on_old)),reduced_prim.minpol);
  [reduced_prim,reduced_first,reduced_second];
}
  


orthogonal_diagonalize(M) =
{
/*Assuming M is already initialized into "algebraic number, polymod" notation*/
  local(N1 = matrix(4,1),
        N2 = matrix(4,1),
        N3 = matrix(4,1),
        N4 = matrix(4,1),
        Q = M[2],  /*The matrix of polymods */
        Q_diag = matrix(4,4));

  N1[1,1] = 1;
  N2[2,1] = 1;
  N3[3,1] = 1;
  N4[4,1] = 1;
  
  
  
  N2 = N2 - N1*(N1~*Q*N2)/(N1~*Q*N1);
  N3 = N3 - N1*(N1~*Q*N3)/(N1~*Q*N1) - N2*(N2~*Q*N3)/(N2~*Q*N2);
  N4 = N4 - N1*(N1~*Q*N4)/(N1~*Q*N1) - N2*(N2~*Q*N4)/(N2~*Q*N2)-N3*(N3~*Q*N4)/(N3~*Q*N3);
  
  Q_diag = matdiagonal([(N1~*Q*N1)[1,1], (N2~*Q*N2)[1,1], (N3~*Q*N3)[1,1], (N4~*Q*N4)[1,1]]);
  
  [M[1],Q_diag]
}

invariant_quaternion_algebra(G,tr_field_primitive = 0,M = 0) = 
{
  local(M_diag,V,square1,square2);
  if (tr_field_primitive == 0,tr_field_primitive = invariant_trace_field(G));
  if (M == 0,M = Mmatrix(G));
  if (type(M) <> "t_VEC",M = anmatinit(M)); 
  if(verbose,print("M: ",M));
  M_diag = orthogonal_diagonalize(M);
  if(verbose,print("M diagonalized: ",M_diag));
  square1 = -(M_diag[2])[1,1]*(M_diag[2])[2,2];
  square2 = -(M_diag[2])[1,1]*(M_diag[2])[3,3];
  V = ancompositum(M_diag[1],tr_field_primitive,1);
  new_prim = V[1];
  M_primitive = lift(V[2]);
  square1 = Mod(subst(lift(square1),x,M_primitive),new_prim[1]);
  square2 = Mod(subst(lift(square2),x,M_primitive),new_prim[1]);
  if(verbose, print(square1,square2));
  
  hilbert_symbol_init(new_prim,square1,square2)
}
 



