real_ramification(symbol) = {
  local(my_roots=polroots(symbol.prim.minpol),
        num_roots = matsize(my_roots)[1],
        ramification=vector(num_roots));
  
  for(i=1,num_roots,
    if(close(imag(my_roots[i]),0),
      if(debug,print((subst(lift(symbol.first),x,real(my_roots[i]))),(subst(lift(symbol.second),x,real(my_roots[i])))));
      ramification[i] = (subst(lift(symbol.first),x,real(my_roots[i]))<0) &&
                        (subst(lift(symbol.second),x,real(my_roots[i]))<0),
	ramification[i] = -1));
  return(ramification)
}


finite_ramification(symbol) =
{
  local(field = nfinit(symbol.prim.minpol),
        /* Ramification can only occur over these primes by MR Thm 2.6.6 */
        prime_list = idealfactor(field, 2 * symbol.first * symbol.second),
        num_primes = matsize(prime_list)[1],
        
        /* Express symbol.first and second in terms of the integral basis
           PARI computes for the field */
        first_wrt_zk = nfalgtobasis(field, symbol.first),
        second_wrt_zk = nfalgtobasis(field, symbol.second),
        
        num_ramfied = 0);
        
  /* Overwrite the multiplicity of the prime ideals with information on
     ramification: 0 if non-ramified, 1 if ramified and -1 if test aborted: */
  for(i=1, num_primes,
    prime_list[i,2] = ramified(symbol, first_wrt_zk, second_wrt_zk, field,
                               prime_list[i,1]);
    if(prime_list[i,2]==-1, print("warning: ramification test failed")));

  return(prime_list)
}

ramified(symbol, first_wrt_zk, second_wrt_zk, field, test_prime) =
{
  /* TO DO: Write code to handle case in which test_prime does not divide
            two or the index of the ring of integers in a faster way */
            
  /* Ramifications occurs iff the Hilbert equation a*x^2+b*y^2==1 has a solution
     in the completion with respect to the valuation given by test_prime;
    according to the equivalence of (a) and (e) in MR Thm 2.3.1 */
  if(debug,print("Current test prime",test_prime));
  return(!solve_hilbert_equation(field, first_wrt_zk,
                                second_wrt_zk, test_prime, 0));
}

/* returns [elt*u^m, odd_val]
   where O_k + u O_k is the inverse of the ideal my_prime (u is contained in
   the description of the prime ideal my_prime),
   m is the higest even number less than or equal to the valuation of elt,
   and odd_val is 0 or 1 depending on the parity of the valuation */
remove_squares_of_prime(field, elt, my_prime) =
{
  local(val = nfeltval(field, elt, my_prime),
        fix_val = 2 * floor(val/2),
        new_elt = nfeltmul(field, elt,
                           nfeltpow(field, my_prime[5]/my_prime[1], fix_val)));
  return([new_elt, val%2])
}
        
solve_hilbert_equation(field,first_wrt_zk,second_wrt_zk,my_prime,target_power) =
{
  local(degree = poldegree(field.pol),
        zero_wrt_zk = nfalgtobasis(field, 0),
        one_wrt_zk = nfalgtobasis(field, 1),
        any_odd_valuation = 0, tmp, sol,
        dyadic_ramification = if(my_prime[1]==2, my_prime[3], 0),
        /* To be able to step through cosets mod my_prime, we compute
           a vector [n_1, ..., n_degree] such that any coset has a unique
           representative with coefficients [k_1, ..., k_degree] wrt to
           the integral basis, where 0<= k_i < n_i. Luckily, PARI can
           compute the Hermite normal form of an ideal; the diagonal is
           almost what we need. */
        my_prime_hnf = idealhnf(field, my_prime),
        my_prime_cosets = vector(degree,i,[0,my_prime_hnf[i,i]-1]));
        if(debug,print(my_prime_cosets));

  tmp = remove_squares_of_prime(field, first_wrt_zk, my_prime);
  if(debug,print("Remove squares ",tmp," ",first_wrt_zk));
  first_wrt_zk = tmp[1];
  any_odd_valuation = any_odd_valuation || tmp[2];
        
  tmp = remove_squares_of_prime(field, second_wrt_zk, my_prime);
  if(debug,print("Remove squares ",tmp," ",second_wrt_zk));
  second_wrt_zk = tmp[1];
  any_odd_valuation = any_odd_valuation || tmp[2];

  if(target_power==0, /* choose sufficiently high power automatically */
    target_power = if(dyadic_ramfication != 0,
                      2 * dyadic_ramification + 3,
                      if(any_odd_valuation, 3, 1)));

  sol = lift_solution(field, one_wrt_zk, zero_wrt_zk, zero_wrt_zk, one_wrt_zk,
                      first_wrt_zk, second_wrt_zk, -one_wrt_zk,
                      0, target_power, my_prime, my_prime_cosets,
                      dyadic_ramification);
  if(sol!=0, return(sol));

  if(debug,print("second lift"));
  sol = lift_solution(field, one_wrt_zk, zero_wrt_zk, zero_wrt_zk, one_wrt_zk,
                      second_wrt_zk, -one_wrt_zk, first_wrt_zk, 
                      0, target_power, my_prime, my_prime_cosets,
                      dyadic_ramification);
  if(sol!=0, return(sol));

  if(debug,print("third lift"));
  sol = lift_solution(field, one_wrt_zk,zero_wrt_zk, zero_wrt_zk, one_wrt_zk,
                      -one_wrt_zk, first_wrt_zk, second_wrt_zk, 
                      0, target_power, my_prime, my_prime_cosets,
                      dyadic_ramification);
  if(sol!=0, return(sol))
}

lift_solution(field, sol_mod, x,y,z, a,b,c, initial_power, target_power,\
              my_prime, my_prime_cosets, dyadic_ramification) =
{
  local(commit_power = if(dyadic_ramification,
                         if(initial_power+1 < dyadic_ramification,
                           2*initial_power + 2,
                           initial_power + dyadic_ramification + 1),
                         initial_power + 1),
       degree = poldegree(field.pol),
       x1, y1, /* attempted solutions mod my_prime^commit_power */
       hilbert_equation_lhs, lhs_val, sol);
 forvec(u=my_prime_cosets,
   forvec(v=my_prime_cosets,
      x1 = nfeltmul(field, u~, sol_mod) + x;
      y1 = nfeltmul(field, v~, sol_mod) + y;
      hilbert_equation_lhs = nfeltmul(field, a, nfeltpow(field, x1, 2)) +
                             nfeltmul(field, b, nfeltpow(field, y1, 2)) +
                             nfeltmul(field, c, nfeltpow(field, z, 2));
      lhs_val = if(norml2(hilbert_equation_lhs)==0,
                  target_power,
                  nfeltval(field, hilbert_equation_lhs, my_prime));
      
      if(lhs_val >= target_power, return([x1,y1,z]));
      
      if(lhs_val >= commit_power,
        sol = lift_solution(field, nfeltmul(field, sol_mod, my_prime[2]),
                            x1, y1, z, a, b, c,
                            initial_power+1, target_power,
                            my_prime, my_prime_cosets, dyadic_ramification);
        if(sol != 0, return(sol)))));
  return(0)
}
